﻿using GDI3D;
using GDI3D.Formats;
using GDI3D.Shader;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Text;
using System.Windows.Forms;

namespace Example
{ 
    public partial class frmMain : Form
    {
        Viewport renderer;
        Size targetSize = new Size(300, 300);
        Matrix4 worldTransform = Matrix4.Identity;
        Mesh triangleMesh = null;
        Texture2 texture = new Texture2("bear.bmp");

        List<FileInfo> files;

        List<ShaderPair> shader = new List<ShaderPair>(); 

        public frmMain()
        {
            InitializeComponent();
            SetupViewport();
            LoadShader();
            LoadFiles();
        }

        private void LoadFiles()
        {
            files = new List<FileInfo>(new DirectoryInfo(Environment.CurrentDirectory).EnumerateFiles("*.ply"));
            cbFile.DataSource = files;
        }

        private void LoadShader()
        {
            // Some example shader
            shader.Add(new ShaderPair((new NormalVertexShader() { Viewport = renderer }).Shader, new NormalPixelShader().Shader, "Normal"));
            shader.Add(new ShaderPair((new ProjectiveVertexShader() { Viewport = renderer }).Shader, BasicPixelShader.VertexColour, "VertexColour"));
            shader.Add(new ShaderPair((new ProjectiveVertexShader() { Viewport = renderer }).Shader, (new TexturePixelShader() { Texture=texture }).Shader, "Texture (Bear)"));
            shader.Add(new ShaderPair((new ProjectiveVertexShader() { Viewport = renderer }).Shader, BearPixelShader, "VertexColour + Texture"));
            shader.Add(new ShaderPair((new NormalVertexShader() { Viewport = renderer }).Shader, BearShader2, "VertexColour + Texture + Normal"));
            cbShader.DataSource = shader;
            renderer.VertexShader = shader[0].VertexShader;
            renderer.PixelShader = shader[0].PixelShader;
        }

        private void SetupViewport()
        {
            renderer = new Viewport(new SizeF(1.5f, 1.5f), targetSize, 35, 0.1f, 21);
            Matrix4 m = Matrix4.Identity;
            m.Translate(0, 0, -5);
            renderer.Transform = m;
            renderer.Clear(Color.CornflowerBlue.ToVector3());
        }

        private void pnlTarget_Paint(object sender, PaintEventArgs e)
        {
            renderer.DrawBufferToDC(e.Graphics, new Point(pnlTarget.Width / 2 - targetSize.Width / 2, pnlTarget.Height / 2 - targetSize.Width / 2));
        }

        private void trk_Scroll(object sender, EventArgs e)
        {
            var translation = new Vector3(trkTransX.Value, trkTransY.Value, trkTransZ.Value);
            var rotation = new Vector3(trkRotX.Value, trkRotY.Value, trkRotZ.Value);

            worldTransform = Matrix4.Identity;
            worldTransform.Translate(translation);
            worldTransform.RotateX(rotation.X * MathHelper.DegToRad);
            worldTransform.RotateY(rotation.Y * MathHelper.DegToRad);
            worldTransform.RotateZ(rotation.Z * MathHelper.DegToRad);
            worldTransform.Invert();

            Draw();
        }

        private void btnLoad_Click(object sender, EventArgs e)
        {
            if (cbFile.SelectedIndex == -1)
                return;
            triangleMesh = Mesh.FromPly(files[cbFile.SelectedIndex].FullName);
            renderer.IndexBuffer = triangleMesh.Indices;
            renderer.VertexBuffer = triangleMesh.Vertices;

            Draw();
        }

        private void cbShader_SelectedIndexChanged(object sender, EventArgs e)
        {
            if (triangleMesh == null || cbShader.SelectedIndex == -1)
                return;
            renderer.VertexShader = shader[cbShader.SelectedIndex].VertexShader;
            renderer.PixelShader = shader[cbShader.SelectedIndex].PixelShader;
            Draw();
        }

        private void btnClear_Click(object sender, EventArgs e)
        {
            renderer.Clear(Color.CornflowerBlue.ToVector3());
            pnlTarget.Invalidate();
        }

        private void Draw()
        {
            renderer.Clear(Color.CornflowerBlue.ToVector3());
            renderer.DrawIndexedTriangles(worldTransform);
            pnlTarget.Invalidate();
        }

        #region CustomShader
        bool BearPixelShader(Vertex input, out Vector3 colour)
        {
            Vector2 uv = input.Attribute.Get<Vector2>("TEXCOORD");
            var textureColour = texture.Sample(uv);
            var vertexColour = input.Attribute.Get<Vector3>("COLOUR");
            if (uv == Vector2.Zero)
                colour = vertexColour;
            else
                colour = MathHelper.Clamp(textureColour * vertexColour * 2, Vector3.Zero, Vector3.One);
            return colour != Vector3.Zero; // Discard black pixel
        }
        bool BearShader2(Vertex input, out Vector3 c)
        {
            c = Vector3.Zero;
            Vector3 viewDirection = -input.Attribute.Get<Vector3>("NORMAL0");
            viewDirection.Normalise();
            float view = MathHelper.Clamp(input.Attribute.Get<Vector3>("NORMAL").Normalised.Dot(viewDirection), 0, 1);
            var uv = input.Attribute.Get<Vector2>("TEXCOORD");
            Vector3 tex = texture.Sample(uv);
            Vector3 vert = input.Attribute.Get<Vector3>("COLOUR");
            if (uv == Vector2.Zero)
                c = vert * view;
            else
                c = MathHelper.Clamp(vert * tex * 2 * view, Vector3.Zero, Vector3.One);
            if (c == Vector3.Zero)
                return false;
            return true;
        }
        #endregion
    }

    public class ShaderPair
    {
        public ShaderPair(Func<Vertex, Vertex> vShader, Viewport.PixelShaderDelegate pShader, string name)
        {
            VertexShader = vShader;
            PixelShader = pShader;
            Name = name;
        }
        public Func<Vertex, Vertex> VertexShader { get; set; }
        public Viewport.PixelShaderDelegate PixelShader { get; set; }
        public string Name { get; set; }
        public override string ToString() => Name;
    }
}
