﻿using System;
using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using System.Runtime.InteropServices;
using System.Text;

namespace GDI3D
{
    public enum ResolutionGate { Fill, Overscan };
    public enum CullMode { AntiClockwise, Clockwise }
    
    public class Viewport : IDisposable
    {
        private object syncLock = new object();
        private Bitmap bitmap = null;
        private const int BytesPerPixel = 4;
        private byte[] buffer = null;
        private float[] zBuffer = null;
        public float FocalLength { get; private set; }
        public float Ratio { get; private set; }
        public float FieldOfView { get; private set; }
        public float NearPlane { get; private set; }
        public float FarPlane { get; private set; }
        public ResolutionGate Mode { get; private set; } = ResolutionGate.Fill;
        const float inchTomm = 25.4f;
        public RectangleF Arpeture { get; private set; }
        public Size DeviceSize { get; private set; }
        public Matrix4 Transform { get; set; }
        public Matrix4 Projection { get; private set; }
        public Matrix4 WorldViewProjection { get; private set; }
        public Matrix4 WorldView { get; private set; }
        public Matrix4 WorldViewInverseTransposed { get; private set; }
        public CullMode CullMode { get; set; }

        public Func<Vertex, Vertex> VertexShader { get; set; }
        public delegate bool PixelShaderDelegate(Vertex vIn, out Vector3 colour);
        public PixelShaderDelegate PixelShader { get; set; }

        public IEnumerable<Vertex> VertexBuffer { get; set; }
        public IEnumerable<UInt32> IndexBuffer { get; set; }

        public Viewport(SizeF arpetureSize, Size deviceSize, float focalLen, float nearPlane, float farPlane, ResolutionGate mode = ResolutionGate.Fill)
        {
            Transform = Matrix4.Identity;
            Mode = mode;
            FocalLength = focalLen;
            NearPlane = nearPlane;
            FarPlane = farPlane;
            DeviceSize = deviceSize;
            float top = ((arpetureSize.Height * inchTomm / 2) / focalLen) * nearPlane;
            float right = ((arpetureSize.Width * inchTomm / 2) / focalLen) * nearPlane;
            Arpeture = RectangleF.FromLTRB(-right, top, right, -top);

            FieldOfView = (float)(2 * 180 / Math.PI * Math.Atan((arpetureSize.Width * inchTomm / 2) / focalLen));
            Ratio = arpetureSize.Width / arpetureSize.Height;
            BuildProjection();
            SetupFramebuffer();
        }
        void SetupFramebuffer()
        {
            buffer = new byte[DeviceSize.Width * DeviceSize.Height * BytesPerPixel];
            zBuffer = new float[DeviceSize.Width * DeviceSize.Height];
            bitmap = new Bitmap(DeviceSize.Width, DeviceSize.Height, DeviceSize.Width * BytesPerPixel, System.Drawing.Imaging.PixelFormat.Format32bppRgb, Marshal.UnsafeAddrOfPinnedArrayElement(buffer, 0));
        }
        private void ResetZ()
        {
            for (int i = 0; i < zBuffer.Length; ++i)
                zBuffer[i] = FarPlane;
        }
        public void LookAt(Vector3 from, Vector3 to) => LookAt(from, to, new Vector3(0, 1, 0));
        public void LookAt(Vector3 from, Vector3 to, Vector3 upDirection)
        {
            Vector3 forward = (from - to).Normalised;
            var tmpUp = upDirection.Normalised;
            Vector3 right = tmpUp.Cross(forward);
            Vector3 up = forward.Cross(right);

            Matrix4 transform = Matrix4.Identity;

            transform.M00 = right.X;
            transform.M01 = right.Y;
            transform.M02 = right.Z;
            transform.M10 = up.X;
            transform.M11 = up.Y;
            transform.M12 = up.Z;
            transform.M20 = forward.X;
            transform.M21 = forward.Y;
            transform.M22 = forward.Z;

            transform.M30 = from.X;
            transform.M31 = from.Y;
            transform.M32 = from.Z;
            Transform = transform;
        }
        private void BuildProjection()
        {
            float deviceAspectRatio = DeviceSize.Width / (float)DeviceSize.Height;
            float xscale = 1;
            float yscale = 1;
            if (Mode == ResolutionGate.Fill)
            {
                if (Ratio > deviceAspectRatio)
                    xscale = deviceAspectRatio / Ratio;
                else
                    yscale = Ratio / deviceAspectRatio;
            }
            else
            {
                if (Ratio > deviceAspectRatio)
                    yscale = Ratio / deviceAspectRatio;
                else
                    xscale = deviceAspectRatio / Ratio;
            }
            float top = Arpeture.Top * yscale;
            float right = Arpeture.Right * xscale;
            Arpeture = RectangleF.FromLTRB(-right, top, right, -top);
            Projection = new Matrix4(Arpeture.Left, Arpeture.Right, Arpeture.Top, Arpeture.Bottom, NearPlane, FarPlane);
        }

        #region IDisposable Support
        private bool disposedValue = false;

        protected virtual void Dispose(bool disposing)
        {
            if (!disposedValue)
            {
                if (disposing)
                {
                    bitmap.Dispose();
                }
                buffer = null;
                disposedValue = true;
            }
        }

        public void Dispose()
        {
            Dispose(true);
        }
        #endregion

        private Vector4 Rasterize(Vector4 v)
        {
            return new Vector4((v.X + 1) * 0.5f * DeviceSize.Width, (1 - (v.Y + 1) * 0.5f) * DeviceSize.Height, v.Z, v.W);
        }

        public void Clear(Vector3 colour)
        {
            for(int i = 0; i < buffer.Length; i+=BytesPerPixel)
            {
                buffer[i] = (byte)(colour.Z * 255);
                buffer[i + 1] = (byte)(colour.Y * 255);
                buffer[i + 2] = (byte)(colour.X * 255);
            }
        }
        public void DrawBufferToDC(Graphics g)
        {
            lock(syncLock)
                g.DrawImageUnscaled(bitmap, 0, 0);
        }

        public void DrawBufferToDC(Graphics g, Point p)
        {
            lock (syncLock)
                g.DrawImageUnscaled(bitmap, p);
        }

        public void DrawIndexedTriangles(Matrix4 world)
        {
            lock (syncLock)
                DrawIndexedTrianglesInternal(world);
        }

        private void DrawIndexedTrianglesInternal(Matrix4 worldToPrimitive)
        {
            // Converts primitives from 3D world space to 2D raster space, if needed each primitive gets clipped between farplane and nearplane
            ResetZ();
            WorldView = worldToPrimitive * Transform;
            WorldViewInverseTransposed = WorldView.Inverse.Transposed;
            WorldViewProjection = WorldView * Projection;
            
            for (int i = 0; i < IndexBuffer.Count(); i += 3)
            {
                Vertex v0 = null;
                Vertex v1 = null; 
                Vertex v2 = null;
                if (CullMode == CullMode.AntiClockwise)
                {
                    v0 = VertexShader(new Vertex(VertexBuffer.ElementAt((int)IndexBuffer.ElementAt(i))));
                    v1 = VertexShader(new Vertex(VertexBuffer.ElementAt((int)IndexBuffer.ElementAt(i + 1))));
                    v2 = VertexShader(new Vertex(VertexBuffer.ElementAt((int)IndexBuffer.ElementAt(i + 2))));
                }
                else
                {
                    v0 = VertexShader(new Vertex(VertexBuffer.ElementAt((int)IndexBuffer.ElementAt(i + 2))));
                    v1 = VertexShader(new Vertex(VertexBuffer.ElementAt((int)IndexBuffer.ElementAt(i + 1))));
                    v2 = VertexShader(new Vertex(VertexBuffer.ElementAt((int)IndexBuffer.ElementAt(i))));
                }
                // Now in clip space

                // Out of bounds -> throw away
                if (v0.Position.W <= 0 && v1.Position.W <= 0 && v2.Position.W <= 0)
                    continue;

                if (BoundaryHelper.IsXOutside(v0.Position, v1.Position, v2.Position) ||
                    BoundaryHelper.IsYOutside(v0.Position, v1.Position, v2.Position) ||
                    BoundaryHelper.IsZOutside(v0.Position, v1.Position, v2.Position, out bool clipTriangle)) // Whole triangle outside the frustum -> throw away
                    continue;
                List<Triangle> triangles = new List<Triangle>() { new Triangle(v0, v1, v2) };
                if (clipTriangle)
                {
                    // Clip the triangle and loop through each
                    triangles = SutherlandHodgman.ClipTriangle(v0, v1, v2);
                }

                foreach (var triangle in triangles)
                {
                    Vertex v0C = triangle.V1;
                    Vertex v1C = triangle.V2;
                    Vertex v2C = triangle.V3;

                    v0C.Position = v0C.Position.Homogenised;
                    v1C.Position = v1C.Position.Homogenised;
                    v2C.Position = v2C.Position.Homogenised;
                    // Divide all vertex attribute (v0, v1, v2 must have equal attributes)
                    foreach(var key in v0C.Attribute)
                    {
                        v1C.Attribute[key].Multiply2(v1C.Position.W);
                        v2C.Attribute[key].Multiply2(v2C.Position.W);
                        v0C.Attribute[key].Multiply2(v0C.Position.W);
                    }
                    // Now in NDC space

                    v0C.Position = Rasterize(v0C.Position);
                    v1C.Position = Rasterize(v1C.Position);
                    v2C.Position = Rasterize(v2C.Position);
                    // Now in raster space
                    // Rasterize current triangle
                    RasterizeTriangle(v0C, v1C, v2C);
                }
            }
        }
        private void RasterizeTriangle(Vertex v0, Vertex v1, Vertex v2)
        {
            // Rasterizes triangle, barycentric version

            // Bounding box
            float xMin = MathHelper.Min3(v0.Position.X, v1.Position.X, v2.Position.X);
            float xMax = MathHelper.Max3(v0.Position.X, v1.Position.X, v2.Position.X);
            float yMin = MathHelper.Min3(v0.Position.Y, v1.Position.Y, v2.Position.Y);
            float yMax = MathHelper.Max3(v0.Position.Y, v1.Position.Y, v2.Position.Y);
            // Out of screen, throw away
            if (xMin > DeviceSize.Width - 1 || xMax < 0 || yMin > DeviceSize.Height - 1 || yMax < 0)
                return;
            // Limit to boundaries
            int x0 = (int)Math.Max(0, xMin);
            int x1 = (int)Math.Min(DeviceSize.Width - 1, Math.Floor(xMax));
            int y0 = (int)Math.Max(0, yMin);
            int y1 = (int)Math.Min(DeviceSize.Height - 1, Math.Floor(yMax));

            float area = 1 / MathHelper.Edge(v0.Position, v1.Position, v2.Position);
            Vector4 firstPixel = new Vector4(x0 + 0.5f, y0 + 0.5f, 0, 0); // Centre of the first pixel

            // Precompute edge function
            float w0c = MathHelper.Edge(v1.Position, v2.Position, firstPixel);
            float w0StepX = (v2.Position.Y - v1.Position.Y);
            float w0StepY = -(v2.Position.X - v1.Position.X);
            w0c -= w0StepX;
            float w0Start = w0c;

            float w1c = MathHelper.Edge(v2.Position, v0.Position, firstPixel);
            float w1StepX = (v0.Position.Y - v2.Position.Y);
            float w1StepY = -(v0.Position.X - v2.Position.X);
            w1c -= w1StepX;
            float w1Start = w1c;

            float w2c = MathHelper.Edge(v0.Position, v1.Position, firstPixel);
            float w2StepX = (v1.Position.Y - v0.Position.Y);
            float w2StepY = -(v1.Position.X - v0.Position.X);
            w2c -= w2StepX;
            float w2Start = w2c;

            for(int y = y0; y <= y1; ++y)
            {
                for(int x = x0; x <= x1; ++x)
                {
                    w0c += w0StepX; // Increment X
                    w1c += w1StepX;
                    w2c += w2StepX;

                    float w0 = w0c; // Barycentric coordinates
                    float w1 = w1c;
                    float w2 = w2c;

                    if (!(w0 >= 0 && w1 >= 0 && w2 >= 0))
                        continue; // Pixel isn't within the triangle

                    w0 *= area;
                    w1 *= area;
                    w2 *= area;

                    Vertex result = new Vertex();
                    // Interpolate final vertex
                    result.Position = v0.Position * w0 + v1.Position * w1 + v2.Position * w2;
                    // Divide by W
                    result.Position = result.Position.Homogenised;

                    int sequentialPosition = y * DeviceSize.Width + x;
                    // Depth Test
                    if (result.Position.Z >= zBuffer[sequentialPosition])
                        continue;

                    // Interpolate vertex attribute
                    foreach (var key in v0.Attribute)
                    {
                        IVertexAttribute data = v0.Attribute[key].Multiply(w0).Add(v1.Attribute[key].Multiply(w1).Add(v2.Attribute[key].Multiply(w2))).Multiply(result.Position.W);
                        result.Attribute.Add(key, data);
                    }

                    if (PixelShader(result, out Vector3 finalColour))
                    {
                        // Write Z Buffer if not discarded
                        zBuffer[sequentialPosition] = result.Position.Z;
                        int sequentialPos2 = (BytesPerPixel * y * DeviceSize.Width) + x * BytesPerPixel;
                        buffer[sequentialPos2] = (byte)(finalColour.Z * 255);
                        buffer[sequentialPos2 + 1] = (byte)(finalColour.Y * 255);
                        buffer[sequentialPos2 + 2] = (byte)(finalColour.X * 255);
                    }
                }
                w0Start += w0StepY; // Increment Y
                w0c = w0Start;
                w1Start += w1StepY;
                w1c = w1Start;
                w2Start += w2StepY;
                w2c = w2Start;
            }
        }
    }
}